"""
Function to setup Orginization Data Model
"""
from merakitools import const,lib,utils
import asyncio
import sys
async def setup_org_data_model():
	"""
	Sets up orginization Data model to be use accross different functions
	Returns:

	"""
	if const.appcfg.all_orgs:
		orgs = utils.getOrginizationsAll()
		utils.set_orginization_by_id(orgs)
	elif len(const.appcfg.allow_org_list) != 0:
		orgdb_tasks = [
				utils.getOrginization(org)
				for org in const.appcfg.allow_org_list
		]
		await asyncio.gather(*orgdb_tasks)
	elif len(const.appcfg.allow_org_list_names) != 0:
		orgs = utils.getOrginizationsAll()
		utils.set_orginizations_by_name(orgs)
	else:
		print("Error No Orginizations List Exiting")
		sys.exit()


import time
import asyncio
import sys
import pandas as pd
from datetime import datetime
from merakitools import const, lib, model, utils,mnetutils
def setup_config_data_model():
    
    for tags in const.appcfg.network_tags:
        model.config_manager.update({tags['golden']: model.Tag_Manager(tags['golden'], tags['target'], const.appcfg.switch_port_tags)})
        for tag in const.appcfg.switch_port_tags:
            model.config_manager[tags['golden']].golden_port_config.update({tag: {}})
            model.config_manager[tags['golden']].target_switches.update({tag: {}})
def setup_approved_config():
    golden_tag = const.appcfg.tag_golden
    for tag in model.config_manager[golden_tag].golden_port_config:
        config = {}
        for config_item in model.config_manager[golden_tag].golden_port_config[tag]:
            if config_item in const.appcfg.switch_port_tags[tag]:
                config.update({config_item: model.config_manager[golden_tag].golden_port_config[tag][config_item]})
        
        model.config_manager[golden_tag].approve_golden_config.update({tag: config})





async def device_config():
    for tags in const.appcfg.network_tags:
        start_time = time.perf_counter()
        print(f'{lib.bc.OKBLUE} Starting port configuration sync for Golden-Tag: {tags["golden"]} to Target-Tag: {tags["target"]} at {start_time:0.5f} {lib.bc.ENDC}')
        model.meraki_nets = {}
        model.golden_nets = {}
        model.config_manager = {}
        setup_config_data_model()
        await mnetutils.setup_org_data_model()
        const.appcfg.tag_golden = tags['golden']
        const.appcfg.tag_target = tags['target']
        org_threads = []
        for org_id in model.meraki_nets:
            org_threads.append(utils.Org_device_proccesssor(org_id))
        for orgthread in org_threads:
            orgthread.start()
        for orgthread in org_threads:
            orgthread.join()
        config_threads = []
        setup_approved_config()
        for org_id in model.meraki_nets:
            config_threads.append(utils.Validatedevice(org_id))
        for configthread in config_threads:
            configthread.start()
        for configthread in config_threads:
            configthread.join()
    
        lib.store_cache(tags['golden'],model.config_manager[tags['golden']],'device_config')
        print(f'Total port configuration sync Runtime: '
              f'{(time.perf_counter() - start_time):0.5f} secounds')


"""
Function to get orginization information for either a white list or all orgs
API key has access to
"""
from merakitools import const, model, lib
from deepdiff import DeepHash


def getOrginizationsAll():
	"""
		Builds out all orginization API key has access to with in the meraki
		DashBoard
		Returns: Dict of Organiztion info from Meraki Dashboard
	"""
	orgs = const.meraki_sdk.organizations.getOrganizations()
	return orgs


def set_orginization_by_id(orgs):
	"""
    Setups Meraki Network Model by ID
    Args:
        orgs: ORGS DICT returned from getOrganization function from meraki SDK

    Returns:

    """
	for org in orgs:
		model.meraki_nets[org['id']] = model.ORGDB(org['id'], org['name'])


def set_orginizations_by_name(orgs):
	"""
    Setups Meraki Network Model by Name
    Serchs the dict of orginiation then runs a compair by a hash of the org name
    Args:
        orgs: ORGS DICT returned from getOrganization function from meraki SDK

    Returns:

    """
	for name in const.appcfg.allow_org_list_names:
		for org in orgs:
			if DeepHash(str(name).upper()) == DeepHash(
					str(org['name']).upper()):
				model.meraki_nets[org['id']] = model.ORGDB(org['id'],
				                                           org['name'])


async def getOrginization(org_id):
	"""
		Gets Orginization information based on org_ID
		Args:
			org_id: Org ID from White List in APPlication Configuration
		Returns:
	"""
	info = const.meraki_sdk.organizations.getOrganization(org_id)
	model.meraki_nets[info['id']] = model.ORGDB(info['id'], info['name'])



__title__ = 'Meraki_Tools'
__description__ = 'Meraki Tools for syncing configuration accross orginaztion, backing up networks and checking configuration compliance'
__url__ = 'http://devnet.cisco.com'
__download_url__ = 'https://devnet.cisco.com'
__author__ = 'Josh Lipton'
__author_email__ = 'jolipton@cisco.com'
__copyright__ = "Copyright (c) 2021 Cisco Systems, Inc."
__license__ = "MIT"
__version__='1.117'



# same as compare() but strips out ID/networkID for profiles/group policies etc
"""
Fuctions to Comapare setting and prep items to be comapired
"""
import copy
from deepdiff import DeepHash

def remove_radius(item):
    """
		Removes radius server information that is unique to each network
		Args:
			item: item to check
		Returns: item with informaiton removed
	"""
    # Need to Move into a fucntion with in the product class
    # had to add some logic to pop the "net_id" and "radsecEnabled".
    # 'net_id' is unique and 'radsecEnabled' is beta for openroaming
    if 'radiusServers' in item:
        item['radiusServers'][0].pop('net_id')
        if 'radsecEnabled' in item['radiusServers'][0]:
            item['radiusServers'][0].pop('radsecEnabled')
        if 'radiusAccountingServers' in item:
            item['radiusAccountingServers'][0].pop('net_id')
            if 'radsecEnabled' in item['radiusAccountingServers'][0]:
                item['radiusAccountingServers'][0].pop('radsecEnabled')
    return item


async def soft_compare(item_a, item_b):
    """
		Perpares items that need spefic values removed as they will always be
		different like network ID
		Args:
			item_a: Itam item_a to search
			item_b: Item item_b to search

		Returns: Sends Item to Compare function for final checks
	"""
    # Saving Configuration to a tempary value so we donot modify orgionalconfig
    try:
        temp_a = copy.deepcopy(item_a)
        temp_b = copy.deepcopy(item_b)
        poplist = [
            'net_id', 'networkId', 'groupPolicyId', 'dnsRewrite',
            'adultContentFilteringEnabled', 'enrollmentString'
        ]
        if isinstance(temp_a, list) and isinstance(temp_b, list):
            for item in temp_a:
                for popkey in poplist:
                    if popkey in item:
                        item.pop(popkey)
            for item in temp_b:
                for popkey in poplist:
                    if popkey in item:
                        item.pop(popkey)
        else:
            for popkey in poplist:
                if popkey in temp_a:
                    temp_a.pop(popkey)
                if popkey in temp_b:
                    temp_b.pop(popkey)
        temp_a = remove_radius(temp_a)
        temp_b = remove_radius(temp_b)
        return await compare(temp_a, temp_b)
    except Exception as error:
        return await compare(temp_a,temp_b)


async def compare(item_a, item_b):
    """
		The function will hash the Items then compare if the Hash is  equal
		Args:
			item_a: Item item_a to Hash, item can e a DICT, Object, STR, INT, ETC
			item_b: Item item_b to Hash, item can e a DICT, Object, STR, INT, ETC

		Returns: True if hash equals False off Hash is not equle
	"""
    hash_a = DeepHash(item_a)[item_a]
    hash_b = DeepHash(item_b)[item_b]
    return hash_a == hash_b



"""
Class May be depericated
"""

from merakitools import products
class OVERRIDES(object):
    def __init__(self):
        self.supported_products = products.supported
        self.List = self._product_functions()
    

    def _product_functions(self):
        List = {}
        for product in self.supported_products:
            method = getattr(products, product)
            _module = dir(method)
            update = [_funct.split('_')[1] for _funct in _module if
                                  _funct.startswith('Update_')]
            get = [_funct.split('_')[1] for _funct in _module if
                               _funct.startswith('Get_')]
            t = {product:{'update': update, 'get': get}}
            List.update(t)
        return List


if __name__ == '__main__':
    p = OVERRIDES()
    print(p.List)


"""
Proccess functions to loop through networks with in an orginizaion and
send that o
"""
import asyncio
import random
import time
from merakitools import mnetutils, model, lib, const
from datetime import datetime


async def validate_network_processor(org_id: str, net_id: str, sdk: object):
    """
    ASYNC Function to synclye through configuration dict for network and
    call the clone functions based on netork net_id 
    Args: 
        org_id:Orginization Id 
        net_id: Network Id 
        sdk: Meraki SDK Object

    Returns:

    """
    org_name = model.meraki_nets[org_id].name
    net_name = model.meraki_nets[org_id].networks[net_id].name
    print(
        f'{lib.bc.OKBLUE}{org_name} - {net_name} {lib.bc.OKGREEN}Starting configuration validation...{lib.bc.ENDC}'
    )
    for product in model.meraki_nets[org_id].networks[net_id].dashboard:
        await mnetutils.clone(org_id, net_id, sdk, product)
def _config_tags(device,sdk):
    switch_port_cfg = []
    tags = {"AP": [16,17,18,19,20,21,22,23,24],"Admin": [1,2,3,4,5,6], "Classroom": [7,8,9,10], "IOT":[11,12,13,14,15]}
    for tag in tags:
        for port in tags[tag]:
            switch_port_cfg.append(sdk.batch.switch.updateDeviceSwitchPort(device, str(port), tags=[tag]))
    return switch_port_cfg

async def validate_device_config_processor(org_id, device, sdk):
    golden_tag = const.appcfg.tag_golden
    switch_port_cfg = []
    
    for tag in model.config_manager[golden_tag].approve_golden_config:
        if device in model.config_manager[golden_tag].target_switches[tag].keys():
            for port in  model.config_manager[golden_tag].target_switches[tag][device]:
                for switch_port in model.meraki_nets[org_id].devices[device].config:
                    if port == switch_port['portId']:
                        approved_port_config = lib.build_approved_config(switch_port,tag)
                        if not await lib.compare(approved_port_config,
                                            model.config_manager[golden_tag].approve_golden_config[tag]):
                            switch_port_cfg.append(sdk.batch.switch.updateDeviceSwitchPort(device,
                                                                          switch_port['portId'],
                                                                          **model.config_manager[golden_tag].approve_golden_config[tag]))
                            print(f"{lib.bc.WARNING} Updating-{lib.bc.OKBLUE}Orginization: {model.meraki_nets[org_id].name} "
                                  f"switch:{device}-port:{str(switch_port['portId'])}-port tag:{tag}{lib.bc.ENDC}")
    if len(switch_port_cfg) > 0:
        print(f'Updating Device {device}')
        return switch_port_cfg
    else:
        print(f'Device {device} does not need update')
        return None


async def proccess_network(org_id: str, net_id: str, sdk: object, tags: list):
    """
    Function checkts Tag of network for Master tag then loops through
    configuration in network and saves items into the Class objects
	Args:
        org_id:Orginization Id 
        net_id: Network Id 
        sdk: Meraki SDK Objectt

	Returns:

	"""
    t1 = time.perf_counter()

    is_golden = await check_is_golden(tags, net_id)
    if is_golden:
        net = model.golden_nets[const.appcfg.tag_golden].networks[
            const.appcfg.tag_golden]
        net_name = const.appcfg.tag_golden
        last_change_ts = await lib.last_change(
            model.golden_nets[const.appcfg.tag_golden].change_log, net_id)
        is_changed = await check_last_change(net.change_log_ts, last_change_ts)
        if model.golden_nets[const.appcfg.tag_golden].cached and is_changed:
            print(
                f'{lib.bc.WARNING} Golden Network Tag: {const.appcfg.tag_golden} Starting Configuration Sync at {t1:0.5f} secound {lib.bc.OKBLUE} - Cached Change TS: '
                f'{model.golden_nets[const.appcfg.tag_golden].networks[const.appcfg.tag_golden].change_log_ts} '
                f'Current TS: {str(last_change_ts)} {lib.bc.ENDC}')
            sync = True
        elif model.golden_nets[
                const.appcfg.tag_golden].cached and not is_changed:
            print(
                f'{lib.bc.OKGREEN} Golden Network Tag: {const.appcfg.tag_golden} is CACHED and has not changed  {lib.bc.ENDC}'
            )
            sync = False
        else:
            print(
                f'{lib.bc.OKGREEN}Started Configuration Sync at {t1:0.5f} secound'
                f' Golden Network Tag: {const.appcfg.tag_golden} is not Cached '
            )
            sync = True
    else:
        last_change_ts = await lib.last_change(
            model.meraki_nets[org_id].change_log, net_id)
        net = model.meraki_nets[org_id].networks[net_id]
        net_name = model.meraki_nets[org_id].networks[net_id].name
        is_changed = await check_last_change(net.change_log_ts, last_change_ts)

        if model.meraki_nets[org_id].cached and is_changed:
            print(
                f'{lib.bc.WARNING} Network:{net_name} HAS CHANGED {lib.bc.OKBLUE}-starting configuration sync at {t1:0.5f} secound{lib.bc.OKBLUE}-Cached Change TS:'
                f'{model.meraki_nets[org_id].networks[net_id].change_log_ts} '
                f'Current TS:{str(last_change_ts)} {lib.bc.ENDC}')
            sync = True
        elif model.meraki_nets[org_id].cached and not is_changed:
            print(
                f'{lib.bc.OKGREEN} Network: {net_name} is CACHED and has not changed  {lib.bc.ENDC}'
            )
            sync = False
        else:
            print(
                f'{lib.bc.OKGREEN}Started Configuration Sync at {t1:0.5f} secound'
                f' Network: {net_name} is not Cached ')
            sync = True
    if 'networks' not in net.products:
        net.products.append('networks')
    approvedList = net.supported
    if sync:
        if const.appcfg.tag_target in net.tags or const.appcfg.tag_override:
            for product in net.products:
                if product in approvedList:
                    waiting = random.randrange(0, 5)
                    await asyncio.sleep(waiting)
                    if is_golden:
                        await mnetutils.sync(sdk, const.appcfg.tag_golden,
                                             const.appcfg.tag_golden, product,
                                             is_golden)
                    else:
                        try:
                            await mnetutils.sync(sdk, net.org_id, net.net_id,
                                                 product, is_golden)

                        except Exception as error:
                            if const.appcfg.debug:
                                print(f'Error: {error}')
                    print(
                        f'{lib.bc.OKBLUE}{net_name} Finished syncing - {lib.bc.WARNING}{product} configuration{lib.bc.ENDC}'
                    )
                else:
                    if const.appcfg.debug:
                        print(f"No sync Module for {product}")
            if is_golden:
                model.golden_nets[net_name].networks[
                    net_name].lastsync = datetime.utcnow()
                model.golden_nets[net_name].networks[
                    net_name].syncruntime = time.perf_counter() - t1
            else:

                model.meraki_nets[org_id].networks[
                    net_id].lastsync = datetime.utcnow()
                model.meraki_nets[org_id].networks[
                    net_id].syncruntime = time.perf_counter() - t1
        if is_golden:
            model.golden_nets[const.appcfg.tag_golden].networks[
                const.appcfg.
                tag_golden].change_log_ts = await lib.get_network_last_change_ts(
                    sdk, org_id, net_id)

        print(f'{lib.bc.OKGREEN}Built Config From Network: {net_name} '
              f'Process took: {net.syncruntime:0.5f} secounds{lib.bc.Default}')

    else:
        if const.appcfg.enable_status:
            t = time.perf_counter() - t1
            print(f'Network: {net_name} '
                  f'Skiped Syncing Process took: {t:0.5f} secounds')


async def check_last_change(change, dashboard_change):
    if change != dashboard_change or change is None:
        return True
    else:
        return False


async def check_is_golden(tags, net_id):
    if const.appcfg.tag_golden in tags or net_id == 'L_575334852396597314Z':
        return True
    else:
        return False



import glob
from os.path import basename, dirname, isfile, join

import automodinit

__all__ = []
automodinit.automodinit(__name__, __file__, globals())
del automodinit
modules = glob.glob(join(dirname(__file__), "*.py"))
supported = [ basename(f)[:-3] for f in modules if isfile(f) and not f.endswith('__init__.py')]


"""
Proccess Orginization to find networks with in Org then sends each network
to the Network Proccessor
"""
import asyncio
import logging
import threading
import time
from datetime import datetime
from merakitools import const, lib, model
from .processor import proccess_network


class Orgsyncprocessor(threading.Thread):
    """
    Stages the sync function to run the processor in ASYNC or load cache
     memeory or loads cache into memeory if using
    cache
    """
    def __init__(self, org_id):
        """
        Init of Org Sync Pocessor
        Args:
            org_id:ID or Org to be proccessed
        """
        threading.Thread.__init__(self)
        self.org_id = org_id
        self.change_log = None
    def run(self):
        start_time = time.perf_counter()
        threading.currentThread().setName(model.meraki_nets[self.org_id].name)
        self.change_log = asyncio.run(lib.get_change_log_from_org(self.org_id))
        if const.appcfg.debug:
            print(f'\tOrgName: {model.meraki_nets[self.org_id].name} Sync '
                  f'Started at: {start_time} '
                  f'Thread PID:{threading.currentThread().native_id}')

        nets = asyncio.run(self._get_org_networks())
        
        if const.appcfg.use_cache:
            lib.load_cache(self.org_id, model.meraki_nets[self.org_id], 'autosync')
        has_golden = False
        remove_list = []
        for net in nets:
            remove_list.append(net['id'])
            if (const.appcfg.tag_golden in net['tags'] or net[
                    'id'] == 'L_575334852396597314') and not model.golden_nets[const.appcfg.tag_golden].cached:
                model.golden_nets[const.appcfg.tag_golden].networks[
                    const.appcfg.tag_golden] = model.MNET(net)
                has_golden = True
            elif (const.appcfg.tag_golden in net['tags'] or net[
                    'id'] == 'L_575334852396597314') and  model.golden_nets[const.appcfg.tag_golden].cached:
                has_golden = True
            if (const.appcfg.tag_target in net['tags'] or const.appcfg.tag_override ) and not model.meraki_nets[self.org_id].cached:
                model.meraki_nets[self.org_id].networks.update(
                    {net['id']: model.MNET(net)})
            elif const.appcfg.tag_target in net['tags'] and net['id'] not in model.meraki_nets[self.org_id].networks.keys():
                model.meraki_nets[self.org_id].networks.update(
                        {net['id']: model.MNET(net)})
            elif net['id'] in model.meraki_nets[self.org_id].networks.keys():
                model.meraki_nets[self.org_id].networks[net['id']].tags = net['tags']
        
        for net in model.meraki_nets[self.org_id].networks:
            if net not in remove_list:
                model.meraki_nets[self.org_id].networks.pop(net)

        asyncio.run(self._async_run(has_golden))

        model.meraki_nets[self.org_id].syncruntime = \
         time.perf_counter() - start_time
        model.meraki_nets[self.org_id].lastsync = datetime.utcnow()
        asyncio.run(lib.update_change_log(self.org_id))
        lib.store_cache(self.org_id, model.meraki_nets[self.org_id], 'autosync')
            
      
        print(
            f'{lib.bc.OKBLUE} OrgName: {model.meraki_nets[self.org_id].name} '
            f'{lib.bc.WARNING}Loaded Last Sync Time: {model.meraki_nets[self.org_id].lastsync}{lib.bc.ENDC}'
            )

    async def _async_run(self,has_golden):
        with lib.MerakiAsyncApi() as sdk:
            logger = logging.getLogger('meraki.aio')
            logger.setLevel(logging.WARNING)
            
            model.meraki_nets[self.org_id].change_log = self.change_log
            if has_golden:
                model.golden_nets[const.appcfg.tag_golden].change_log = self.change_log
            
            net_tasks = [
                proccess_network(self.org_id, net, sdk, model.meraki_nets[self.org_id].networks[net].tags)
                for net in model.meraki_nets[self.org_id].networks
            ]
            await asyncio.gather(*net_tasks)


    async def _get_org_networks(self):
        with lib.MerakiAsyncApi() as sdk:
            return await sdk.organizations.getOrganizationNetworks(self.org_id)


"""
Preps function to poccess clone in Async
"""
import asyncio
import logging
import threading
import random
from merakitools import const, lib, model
from .processor import validate_device_config_processor, _config_tags
from meraki.exceptions import AsyncAPIError

class Validatedevice(threading.Thread):
    """
    Setups the orginization objet to run through clone and validate proccess
    """
    
    def __init__(self, org_id):
        """
        Init Function or Validateorginization
        Args:
            org_id(string): Orginization ID from Meraki
        """
        threading.Thread.__init__(self)
        self.org_id = org_id
        self.name = model.meraki_nets[org_id].name
        self.change_log = []
    
    def run(self):
        """
        Start the ASYNC fun fuctions and waits for completions to restore
        Cache
        Returns:

        """
        asyncio.run(self._async_run())
        lib.store_cache(self.org_id, model.meraki_nets[self.org_id],
                        'autosync')
    
    async def _async_run(self):
        """
        Async version of the run function loops through all networks
        then sends the network to the valate proccessor
        Returns:

        """
        print(
                f'\tOrgName: {self.name}Thread PID:{threading.currentThread().native_id}'
        )
        threading.currentThread().setName(self.name)
        print(f'\tThread Name:{threading.currentThread().name}')
        device_configs = []
        with lib.MerakiAsyncApi() as sdk:
            logger = logging.getLogger('meraki.aio')
            logger.setLevel(logging.WARNING)
            for device in model.meraki_nets[self.org_id].devices:
                # device_configs.append(validate_device_config_processor(self.org_id, device, sdk))
                if model.meraki_nets[self.org_id].devices[device].product == 'switch':
                    device_configs.append(
                            await validate_device_config_processor(self.org_id,device, sdk))
        
        all_configs = (await self._create_all_config(device_configs))
        
        batch_configs = await self._create_batch_configs(all_configs)
        with lib.MerakiAsyncApi() as sdk:
            if batch_configs is not None:
                for config in batch_configs:
                    try:
                        result = await sdk.organizations.createOrganizationActionBatch(
                            self.org_id, batch_configs[config], confirmed=True)
                        status = await sdk.organizations.getOrganizationActionBatch(
                            self.org_id, result['id'])

                        while not status['status']['completed'] or \
                                status['status']['failed']:
                            delay = random.randrange(5, 10)
                            await asyncio.sleep(delay)
                            try:
                                status = await sdk.organizations.getOrganizationActionBatch(
                                    self.org_id, result['id'])
                                if const.appcfg.enable_status:
                                    print(
                                        f"Orginization Name: {self.name} Action Batch ID: {result['id']} statuss: {str(status['status'])}")
                            except AsyncAPIError as error:
                                if const.appcfg.debug:
                                    print(f"Error runing Action Batch: {str(error)}")
                            
                        print(
                                f"f'{lib.bc.OKBLUE}Orginization: {self.name} - {lib.bc.WARNING} Action Batch ID: {result['id']} Batch: {(int(config) + 1 )} of {len(batch_configs)} "
                                f"{lib.bc.OKGREEN}Status Completed: :{status['status']['completed']} failed: {status['status']['failed']}{lib.bc.ENDC}")
                        if status['status']['failed']:
                            print(f'Failed Config')
                    except AsyncAPIError as error:
                        if const.appcfg.debug:
                            print(f"Error runing Action Batch: {str(error)}")
    
    async def _create_all_config(self, net_configs):
        all_configs = []
        for device_configs in net_configs:
            if device_configs is not None:
                for config in device_configs:
                    all_configs.append(config)
        return all_configs
    
    async def _create_batch_configs(self, all_configs):
        if len(all_configs) <= 100:
            return {0: all_configs}
        else:
            return_config_dict = {}
            return_config = []
            count = 0
            for config in all_configs:
                if len(return_config) <= 99:
                    return_config.append(config)
                else:
                    return_config_dict.update({count: return_config})
                    count += 1
                    return_config = None
                    return_config = [config]
            return_config_dict.update({count: return_config})
            return return_config_dict
    # validate_device_config_processor(self.org_id,device, sdk)
    # net_compare_task = [
    #		_config_tags(self.org_id,device,sdk)
    #		for device in model.meraki_nets[self.org_id].devices]
    #
    # await asyncio.gather(*net_compare_task)
